__author__ = "Markus Pichler"
__copyright__ = "Copyright 2021, University of Technology Graz"
__credits__ = ["Markus Pichler"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Markus Pichler"

from pandas import read_fwf, IndexSlice as idx
from os import path
from io import BytesIO, TextIOWrapper

from scipy.interpolate import interp2d

from ehyd_tools.in_out import FIELDS, DATA_KIND, _get_request


class INDICES:
    DURATION = 'duration'
    RETURN_PERIOD = 'return period'
    CALCULATION_METHOD = 'calculation method'


class INDICES_GER:
    DURATION = 'Dauerstufe'
    RETURN_PERIOD = 'Jährlichkeit'
    CALCULATION_METHOD = 'Berechnungstyp'


def get_ehyd_design_rainfall_file(grid_point_number=5214):
    r = _get_request(grid_point_number, data_kind=DATA_KIND.DESIGN_PRECIPITATION, field=None, file_number=None)
    return TextIOWrapper(BytesIO(r.content), encoding='ISO-8859-1')


def save_ehyd_design_rainfall_pdf(grid_point_number, fn):
    fn = path.join(fn, 'Bemessungsniederschlag_Gitterpunkt_{no}.pdf'.format(no=grid_point_number))
    if not path.isfile(fn):
        r = _get_request(grid_point_number, data_kind=DATA_KIND.DESIGN_PRECIPITATION, field=FIELDS.PDF, file_number=None)
        open(fn, 'wb').write(r.content)


def read_ehyd_design_rainfall(filepath_or_buffer):
    if isinstance(filepath_or_buffer, str):
        txt_file = open(filepath_or_buffer, 'r', encoding='ISO-8859-1')
    else:
        txt_file = filepath_or_buffer

    # The first three lines are purely informal and do not belong to the table.
    first_three_lines = [txt_file.readlines(1) for _ in range(3)]
    # print(*first_three_lines, sep='\n')

    df = read_fwf(txt_file, header=0, skiprows=0, index_col=[0, 1, 2],
                  encoding='ISO-8859-1',
                  comment='-',
                  skip_blank_lines=True)
    txt_file.close()

    # drop empty rows
    df = df.dropna(axis=0, how='all')

    # convert string column names to integers
    df.columns = [int(c.replace('T', '')) for c in df.columns]
    # df.columns.name = 'Jährlichkeit'
    df.columns.name = INDICES.RETURN_PERIOD

    # index "0" (="DAUER") is just the string representative of the index "duration"
    df.index = df.index.droplevel(0)
    # originally named ["DAUERMIN", "TYP"]
    # df.index.names = ['Dauerstufe', 'Berechnungstyp']
    df.index.names = [INDICES.DURATION, INDICES.CALCULATION_METHOD]

    # set duration index as integer
    df.index = df.index.set_levels([df.index.levels[0].astype(int), df.index.levels[1]])

    return df


def get_ehyd_design_rainfall(grid_point_number):
    return read_ehyd_design_rainfall(get_ehyd_design_rainfall_file(grid_point_number))


def get_max_calculation_method(df, methods=None):
    if methods is None:
        methods = ['ÖKOSTRA', 'Bemessung']

    return df.loc[idx[:, methods], :].groupby(axis=0, level=0).max().copy()


def get_calculation_method(df, method='Bemessung'):
    return df.xs(method, axis=0, level=INDICES.CALCULATION_METHOD, drop_level=True).copy()


def get_rainfall_height(table, return_period, duration):
    f = interp2d(x=table.columns.astype(float).values, y=table.index.astype(float).values, z=table.values, kind='linear')
    return float(f(return_period, duration)[0])



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

"""
execute the command line tool
"""

from .cl_tool import execute_cl_tool

execute_cl_tool()



__author__ = "David Camhy, Markus Pichler"
__copyright__ = "Copyright 2017, University of Technology Graz"
__credits__ = ["David Camhy", "Markus Pichler"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "David Camhy, Markus Pichler"

import codecs
import json
import warnings
import io
import os
import re
from zipfile import ZipFile

import pandas as pd
import requests
from pandas.errors import ParserError

from .sww_utils import guess_freq

ENCODING = 'iso8859'


def csv_args(unix=False):
    """
    how to read and write csv file
    differs between windows and unix (mac+linux)

    Args:
        unix (bool): if it is a unix computer

    Returns:
        dict: arguments to read and write csv files
    """
    if unix:
        return dict(sep=',', decimal='.')
    else:
        return dict(sep=';', decimal=',')


def check_path(pth=None):
    """
    use the local directory if not path is give

    Args:
        pth (str): path to directory

    Returns:
        str: path to directory
    """
    if pth == '':
        return pth
    elif pth is None:
        return ''
    elif os.path.isdir(pth):
        return pth
    else:
        raise UserWarning('Path is not available')


PARQUET_ERROR = ModuleNotFoundError("""Error: Unable to find a usable engine to read or write parquet files.
A suitable version of pyarrow (recommended) or fastparquet (alternative) is required for parquet support.
Use pip or conda to install:
- pyarrow (https://pypi.org/project/pyarrow/) or 
- fastparquet (https://pypi.org/project/fastparquet/).""")


def export_series(series, filename, export_path=None, save_as='csv', unix=False):
    """
    export the series to a given format
    may be extended

    Args:
        series (pandas.Series):
        filename (str): name of the file
        export_path (str): path where the file will be stored.
        save_as (str): export format
        unix (bool): whether to use "," or ";" for the csv

    Returns:
        str: path to created file
    """
    fn = os.path.join(check_path(export_path), '{}.{}'.format(filename, save_as))

    if save_as == 'csv':
        series.to_csv(fn, **csv_args(unix))

    elif save_as == 'parquet':
        try:
            series.to_frame().to_parquet(fn)
        except ImportError:
            raise PARQUET_ERROR

    else:
        raise NotImplementedError('Sorry, but only csv files are implemented. Maybe there will be more options soon.')

    return fn


def import_series(filename, series_label='precipitation', index_label='datetime', unix=False):
    """

    Args:
        filename (str):
        series_label (str):
        index_label (str):
        unix (bool): whether to use "," or ";" for the csv

    Returns:
        pandas.Series: rain series
    """
    if filename.endswith('csv'):
        try:
            ts = pd.read_csv(filename, index_col=0, header=0, squeeze=True, **csv_args(unix))
            ts.index = pd.to_datetime(ts.index)
            ts.name = series_label
            ts.index.name = index_label
            return ts
        except (ParserError, UnicodeDecodeError):
            return read_ehyd_file(filename)
    elif filename.endswith('parquet'):
        try:
            return pd.read_parquet(filename).iloc[:, 0].asfreq('T').copy()
        except ImportError:
            raise PARQUET_ERROR
    else:
        raise NotImplementedError('Sorry, but only csv files are implemented. Maybe there will be more options soon.')


# ######################################################################################################################
STATIONS_PRECIPITATION_HIGH_RES = json.load(open(os.path.join(os.path.dirname(__file__), 'ehyd_stations.json'), 'r', encoding='utf-8'))
"""Niederschlagsstationen mit Langzeitserie mit Minutensummen"""


def get_high_res_station(identifier):
    """
    get the name of the station based on the id number

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station

    Returns:
        str: name of the station
    """
    return STATIONS_PRECIPITATION_HIGH_RES[identifier]


class DATA_KIND:
    MEASUREMENT = 'MessstellenExtraData'
    DESIGN_PRECIPITATION = 'BemessungsniederschlagExtraData'


class FIELDS:
    NIEDERSCHLAG = 'nlv'
    QUELLEN = 'qu'
    GRUNDWASSER = 'gw'
    OBERFLAECHENWASSER = 'owf'
    PDF = 'pdf'


_path_file = os.path.dirname(__file__)
_stations_files = {FIELDS.NIEDERSCHLAG: 'niederschl_lufttemp_verdunst.csv',
                   FIELDS.QUELLEN: 'unteririsches_wasser.csv',
                   FIELDS.OBERFLAECHENWASSER: 'oberflaechenwasser.csv'}


def get_ehyd_station_frame(field):
    return pd.read_csv(os.path.join(_path_file, _stations_files[field]), index_col=0, header=0)


EHYD_STATIONS = {k: get_ehyd_station_frame(k).to_dict(orient='index') for k, v in _stations_files.items()}


def get_ehyd_stations(field=FIELDS.NIEDERSCHLAG):
    """
    Stationsinformationen wie: Mst. ID,Messstellen Name,Jahr,Bundesland,Flussgebiet,Seehöhe

    Args:
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)

    Returns:
        dict: Stationsinformationen
    """
    return EHYD_STATIONS[field]


def get_basic_station_meta(identifier, field=FIELDS.NIEDERSCHLAG):
    """
    Stationsinformationen wie: Mst. ID,Messstellen Name,Jahr,Bundesland,Flussgebiet,Seehöhe

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)

    Returns:
        dict: Stationsinformationen
    """
    return get_ehyd_stations(field)[identifier]


def _get_url(identifier, data_kind=DATA_KIND.MEASUREMENT, field=FIELDS.NIEDERSCHLAG, file_number=2):
    """
    get the URL to the specific file

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)
        file_number: 1 - ... (see available files with function `available_files`)

    Example-URLS:
        - https://ehyd.gv.at/eHYD/MessstellenExtraData/qu?id=395293&file=1
        - https://ehyd.gv.at/eHYD/MessstellenExtraData/nlv?id=101063&file=1
        - https://ehyd.gv.at/eHYD/MessstellenExtraData/gw?id=325274&file=1
        - https://ehyd.gv.at/eHYD/MessstellenExtraData/owf?id=211912&file=2
        - https://ehyd.gv.at/eHYD/BemessungsniederschlagExtraData?id=4108
        - https://ehyd.gv.at/eHYD/BemessungsniederschlagExtraData/pdf?id=3499

    Returns:
        str: url to the file
    """
    url = f'https://ehyd.gv.at/eHYD/{data_kind}'
    if field:
        url += f'/{field}'
    url += f'?id={identifier}'
    if file_number:
        url += f'&file={file_number}'
    return url


_REQUESTS = dict()


def _get_request(identifier, data_kind=DATA_KIND.MEASUREMENT, field=FIELDS.NIEDERSCHLAG, file_number=1) -> requests.Response:
    """get request of website"""
    url = _get_url(identifier=identifier, field=field, file_number=file_number, data_kind=data_kind)
    if url not in _REQUESTS:
        _REQUESTS[url] = requests.get(url, allow_redirects=True)
    return _REQUESTS[url]


def _file_available(r: requests.Response) -> bool:
    """if any file is available in give request"""
    return 'content-disposition' in r.headers


def _get_filename(r: requests.Response) -> str:
    """get filename of request"""
    # h['content-disposition'] = 'attachment; filename=N-Minutensummen-112086.zip'
    return r.headers['content-disposition'].split('filename=')[1]


def available_files(identifier, field=FIELDS.NIEDERSCHLAG, data_kind=DATA_KIND.MEASUREMENT):
    """
    get the file of the series of the station

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)

    Returns:
        dict[int, str]: dictionary of {file-number: file-name}
    """
    files = dict()
    for file_number in range(1, 15):
        r = _get_request(identifier=identifier, field=field, file_number=file_number, data_kind=data_kind)
        if _file_available(r):
            files[file_number] = _get_filename(r)
        else:
            break
    return files


def _get_file_from_request(r: requests.Response) -> [str, (io.TextIOWrapper or io.IOBase)]:
    filename = _get_filename(r)
    if ('N-Minutensummen' in filename) and ('.zip' in filename):
        c = r.content
        z = ZipFile(io.BytesIO(c))
        filename = z.namelist()[0]
        csv_file = io.TextIOWrapper(z.open(filename), encoding=ENCODING)
        return csv_file, filename.split('.')[0]
    elif ('.csv' in filename) or ('.txt' in filename):
        csv_file = io.TextIOWrapper(io.BytesIO(r.content), encoding=ENCODING)
        return csv_file, filename.split('.')[0]
    else:
        return None, None


def _get_file(identifier, field=FIELDS.NIEDERSCHLAG, file_number=1, data_kind=DATA_KIND.MEASUREMENT):
    """
    get the file of the series of the station

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)
        file_number (int): file-number (>= 1)

    Returns:
        ((io.TextIOWrapper | io.IOBase), str): data file
    """
    r = _get_request(identifier=identifier, field=field, file_number=file_number, data_kind=data_kind)
    if _file_available(r):
        csv_file, filename = _get_file_from_request(r)
        if csv_file is None:
            raise NotImplementedError('This kind of request is not implemented (yet?). Sorry!')
        return csv_file, filename


def get_ehyd_files(identifier, field=FIELDS.NIEDERSCHLAG, data_kind=DATA_KIND.MEASUREMENT):
    """
    get the files of the series of one station

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)

    Returns:
        dict: with key=filename and value=tuple(meta_data, timeseries_data)
    """
    files = dict()
    for file_number, filename in available_files(identifier, field=field, data_kind=data_kind).items():
        if file_number == 1:
            files[filename] = get_station_reference_data(identifier, field=field, data_kind=data_kind)
        else:
            r = _get_request(identifier=identifier, field=field, file_number=file_number, data_kind=data_kind)
            files[filename] = read_ehyd_file(*_get_file_from_request(r))
    return files


def _parse_meta_data(meta_str):
    # print('#' * 100)
    # print(meta_str)
    meta = dict(_raw=meta_str)
    currant_table = None
    currant_header = None
    table_key = None
    is_table = False
    sep = re.compile(r':\s+')
    lines = iter(meta_str.split('\n'))
    for line in lines:
        if not is_table and line.endswith(':'):
            # start first table
            is_table = True
            table_key = line[:-1].split(':  ')[0]
            currant_header = re.split(sep, line.strip().strip(':'))
            currant_table = list()

        elif is_table and not line.startswith(' ') and line.endswith(':'):
            # end table | start new table
            meta[table_key] = currant_table
            # ------------
            table_key = line[:-1].split(':  ')[0]
            currant_header = re.split(sep, line.strip().strip(':'))
            currant_table = list()

        elif is_table and line.startswith(' ') and line.endswith(':'):
            # line is header
            currant_header = re.split(sep, line.strip().strip(':'))

        elif is_table and line.startswith(' '):
            # values in table
            values = re.split(r'\s\s+', line.strip())
            currant_table.append(dict(zip(currant_header, values)))

        elif ':' in line:
            # simple key: value | end table here
            key, *value = re.split(sep, line)
            value = ': '.join(value)
            if key in meta:
                if isinstance(meta[key], str):
                    meta[key] = [meta[key], value]
                elif isinstance(meta[key], list):
                    meta[key].append(value)
            else:
                meta[key] = value
            if is_table:
                meta[table_key] = currant_table
                is_table = False

        elif line.strip() == '':
            # empty line | end table
            pass

        else:
            value = line.strip()
            if key in meta:
                if isinstance(meta[key], str):
                    meta[key] = [meta[key], value]
                elif isinstance(meta[key], list):
                    meta[key].append(value)
            else:
                meta[key] = value
            # print('UNKOWN:', line)

    if table_key not in meta:
        meta[table_key] = currant_table

    # print(json.dumps(meta, indent=4, ensure_ascii=False))
    return meta


def get_station_reference_data(identifier, field=FIELDS.NIEDERSCHLAG, data_kind=DATA_KIND.MEASUREMENT):
    """
    get the station reference data (=Stammdaten der Station)

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)

    Returns:
        dict: meta data of station
    """
    # parse_meta_data(_get_file(identifier=identifier, field=field, file_number=1, data_kind=data_kind).read())
    return _parse_meta_data(_get_file(identifier=identifier, field=field, file_number=1, data_kind=data_kind)[0].read())


def _split_file(file):
    """
    split the file in meta-data and time-series-data

    Args:
        file (io.IOBase): file of the series

    Returns:
        (list, list): meta-data and time-series-data
    """
    lines = file.readlines()
    file.close()

    i = lines.index('Werte:\n')

    meta = lines[:i]
    data = lines[i+1:]
    return meta, data


def read_ehyd_file(filepath_or_buffer, series_label='data'):
    """
    read ehyd data file

    Args:
        filepath_or_buffer (io.IOBase | str):
        series_label (str): name of the series

    Returns:
        pandas.Series: time-series with meta-data ts.attrs
    """
    if isinstance(filepath_or_buffer, str):
        csv_file = codecs.open(filepath_or_buffer, 'r', encoding=ENCODING)
    elif isinstance(filepath_or_buffer, io.IOBase):
        csv_file = filepath_or_buffer
    else:
        raise NotImplementedError()

    # ___________________________
    meta, data = _split_file(csv_file)

    # ___________________________
    ts = pd.read_csv(io.StringIO('\n'.join(data).replace(' ', '')), sep=';', decimal=',', index_col=0,
                     na_values='Lücke', header=None, squeeze=True,  # names=[series_label],
                     date_parser=lambda s: pd.to_datetime(s, format='%d.%m.%Y%H:%M:%S')
                     )

    if isinstance(ts, pd.DataFrame):
        ts = ts.dropna(axis=1, how='all')

        if ts.columns.size == 1:
            ts = ts.iloc[:, 0].copy()

    if isinstance(ts, pd.Series):
        ts.name = series_label

    # ___________________________
    ts.index.name = 'datetime'
    # ts = ts.rename_axis(index_label, axis='index')

    freq = guess_freq(ts.index)

    if isinstance(freq, pd.tseries.offsets.MonthBegin) or (freq <= pd.Timedelta(days=7)):
        ts = ts.resample(freq).ffill()

    # ___________________________
    ts.attrs = _parse_meta_data(''.join(meta).replace(';', ''))
    # meta = pd.Series(meta).str.replace('\n', '').str.split(';', expand=True).fillna('').apply(lambda x: x.str.strip())
    return ts


def get_ehyd_data(identifier, field=FIELDS.NIEDERSCHLAG, file_number=2, data_kind=DATA_KIND.MEASUREMENT):
    """
    get the series from the ehyd platform

    Args:
        identifier (int): Gitterpunktnummer or HZBNR of the station
        field (str): nlv; qu; gw; owf (use constant struct: `FIELDS`)
        data_kind (str): MessstellenExtraData; BemessungsniederschlagExtraData (use constant struct: `DATA_KIND`)
        file_number (int): file-number (>= 1)

    Returns:
        pandas.Series: time-series with meta-data ts.attrs
    """
    if identifier not in get_ehyd_stations(field):
        raise ValueError(f'Identifier "{identifier}" not in ehyd!')

    return read_ehyd_file(*_get_file(identifier=identifier, field=field, file_number=file_number, data_kind=data_kind))



__author__ = "David Camhy, Markus Pichler"
__copyright__ = "Copyright 2018, University of Technology Graz"
__credits__ = ["David Camhy", "Markus Pichler"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "David Camhy, Markus Pichler"

from pandas.tseries.frequencies import to_offset
import pandas as pd

START = 'start'
END = 'end'


def _index_series(date_time_index):
    """create a time series from a datetime index without loosing the timezone info"""
    if isinstance(date_time_index, pd.DatetimeIndex) and date_time_index.tzinfo is not None:
        return pd.Series(data=date_time_index, index=date_time_index)
    else:
        return date_time_index.to_series()


def span_table(span_bool, min_span=None):
    """
    time span with consist "True" with a minimum span of <min_span> are the resulting events

    Args:
        span_bool (pandas.Series[bool]): "True"=Event
        min_span (pandas.Timedelta): minimum time range of an event (default=freq of index)

    Returns:
        pandas.DataFrame: with the columns:
            'start' = start-time,
            'end' = end-time,
    """
    index = span_bool.index

    # minimum duration which is considered as one event
    if min_span is None:
        min_span = guess_freq(index).delta  # TODO: Will not work on monthly or yearly data steps

    # pandas.Series with DatetimeIndex as index AND data
    temp = _index_series(index[span_bool])

    # first value in diff will default to NaN
    # fill value is set to double the value of the greater than operation = fixed true value
    start_bool = temp.diff().gt(min_span, fill_value=min_span * 2)
    end_bool = start_bool.shift(-1).fillna(True)

    events = pd.DataFrame()

    events[START] = temp[start_bool].index
    events[END] = temp[end_bool].index

    return events


def guess_freq(date_time_index, default=pd.Timedelta(minutes=1)):
    """
    guess the frequency by evaluating the most often frequency

    Args:
        date_time_index (pandas.DatetimeIndex): index of a time-series
        default (pandas.Timedelta):

    Returns:
        DateOffset: frequency of the date-time-index
    """
    freq = date_time_index.freq
    if pd.notnull(freq):
        return to_offset(freq)

    if not len(date_time_index) <= 3:
        freq = pd.infer_freq(date_time_index)  # 'T'

        if pd.notnull(freq):
            return to_offset(freq)

        delta_series = date_time_index.to_series().diff(periods=1).fillna(method='backfill')
        counts = delta_series.value_counts()
        counts.drop(pd.Timedelta(minutes=0), errors='ignore')

        if counts.empty:
            delta = default
        else:
            delta = counts.index[0]
            if delta == pd.Timedelta(minutes=0):
                delta = default
    else:
        delta = default

    return to_offset(delta)



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import argparse


def ehyd_arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-id',
                        help='the id number for the station from the ehyd.gv.at platform',
                        required=False, type=int)
    parser.add_argument('--input',
                        help='path to the rain input file including the filename',
                        required=False)

    parser.add_argument('--max10a',
                        help='consider only 10 years with the most availability (for clipping the data)',
                        required=False, action='store_true')

    parser.add_argument('--start',
                        help='custom start time (Format="YYYY-MM-DD") for clipping the data',
                        required=False)

    parser.add_argument('--end',
                        help='custom end time (Format="YYYY-MM-DD") for clipping the data',
                        required=False)

    parser.add_argument('--add_gaps',
                        help='save a gaps-table as a csv-file',
                        required=False, action='store_true')

    parser.add_argument('--to_csv',
                        help='save the time-series as csv-file '
                             '(to the current directory if the id is used or in the directory of the input-file)',
                        required=False, action='store_true')

    parser.add_argument('--to_parquet',
                        help='save the time-series as parquet-file '
                             '(to the current directory if the id is used or in the directory of the input-file)'
                             ' - parquet is a much faster as csv to read and write',
                        required=False, action='store_true')

    parser.add_argument('--plot',
                        help='save a bar-plot with monthly sums and availability as a png-file',
                        required=False, action='store_true')

    parser.add_argument('--statistics',
                        help='save the basic statistics (sum, max & min) as a txt-file',
                        required=False, action='store_true')

    parser.add_argument('--meta',
                        help='save the meta-data presented in ehyd as a txt-file',
                        required=False, action='store_true')

    parser.add_argument('--unix',
                        help='export the csv files with a "," as separator and a "." as decimal sign '
                             '(otherwise ";" as separator and a "," as decimal sign will be used)',
                        required=False, action='store_true')

    return parser.parse_args()



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from pandas import to_datetime, DataFrame, Series
from os import path
import json
from webbrowser import open as show_file
from .data_processing import (data_validation, data_availability, max_10a, check_period, rain_plot, create_statistics,
                              start_end_date, )
from .arg_parser import ehyd_arg_parser
from .in_out import (get_ehyd_data, import_series, export_series, csv_args, get_station_reference_data, FIELDS,
                     get_basic_station_meta, DATA_KIND, )
from .sww_utils import span_table


def convert_time(time=None, helper=''):
    try:
        return to_datetime(time)
    except ValueError():
        raise UserWarning('Wrong time format for the {} time. Use the format="YYYY-MM-DD"'.format(helper))


def output_filename(fn):
    if '/' in fn or '\\' in fn:
        message = 'in the file "{}"'
    else:
        message = 'in the current working directory as "{}"'

    return message.format(fn)


def execute_cl_tool():
    """execute the command line tool"""
    # parse command line arguments
    args = ehyd_arg_parser()
    # convert argparse object to dict
    args_dict = vars(args).copy()
    args_dict['id_'] = args.id
    del args_dict['id']
    args_dict['input_'] = args.input
    del args_dict['input']
    run_script(**args_dict)


def get_data(id_=None, input_=None, meta=False, unix=False):
    """
    get time-series data with its name

    Args:
        id_ (str | int): ID-number of the series to be downloaded
        input_ (str): filename of the series to be imported
        meta (bool): if the meta data should be saved as txt file
        unix (bool): if the <input_> file is a csv-file: True={sep=',', decimal='.'} False={sep=';', decimal=','}

    Returns:
        (pandas.Series, str): precipitation series and the label of the series
    """
    if id_ is not None:
        id_number = id_
        name = 'ehyd_{}'.format(id_number)

        print(f'You choose the id: "{id_number}" with the meta-data: {get_basic_station_meta(id_number, field=FIELDS.NIEDERSCHLAG)}.')

        if meta:
            with open('{}_meta.json'.format(name), 'w') as f:
                json.dump(get_station_reference_data(id_number, field=FIELDS.NIEDERSCHLAG, data_kind=DATA_KIND.MEASUREMENT), f, indent=4)
                print('The meta-data are saved in {}.'.format(output_filename(f.name)))

        series = get_ehyd_data(id_number, field=FIELDS.NIEDERSCHLAG, data_kind=DATA_KIND.MEASUREMENT)

    # __________________________________________________________________________________________________________________
    elif input_ is not None:
        fn = input_
        series = import_series(fn, unix=unix)
        print('The data in "{}" were imported and are used as the precipitation time-series.'.format(fn))
        name = path.basename(fn).split('.')[0]

    # __________________________________________________________________________________________________________________
    else:
        raise UserWarning('No data selected. Use either a input file or a id. See help menu.')

    print('The selected time-series starts at "{:%Y-%m-%d}" and ends at "{:%Y-%m-%d}".'.format(*start_end_date(series)))

    return series, name


def run_script(start=None, end=None, id_=None, input_=None, meta=False, max10a=False, add_gaps=True, statistics=False,
               to_csv=False, to_parquet=False, plot=False, unix=False):
    # __________________________________________________________________________________________________________________
    if start is not None:
        start = convert_time(start, 'start')

    # __________________________________________________________________________________________________________________
    if end is not None:
        end = convert_time(end, 'end')

    # __________________________________________________________________________________________________________________
    series, name = get_data(id_, input_, meta, unix)

    # __________________________________________________________________________________________________________________
    series = series[start:end].copy()
    check_period(series)

    # ______________________________________________________
    tags = None
    availability = None

    # __________________________________________________________________________________________________________________
    if max10a:
        tags = data_validation(series)
        availability = data_availability(tags)
        start, end = max_10a(availability)
        series = series.loc[start:end].copy()
        check_period(series)

    if start or end or max10a:
        print('The time-series is clipped to '
              'start at "{:%Y-%m-%d}" and end at "{:%Y-%m-%d}".'.format(*start_end_date(series)))

    # __________________________________________________________________________________________________________________
    if add_gaps:
        if tags is None:
            tags = data_validation(series)
        if availability is None:
            availability = data_availability(tags)

        gaps = span_table(~availability)
        gaps['delta'] = (gaps['end'] - gaps['start']).dt.total_seconds() / (60 * 60 * 24)
        gaps = gaps.sort_values('delta', ascending=False)
        gaps.columns = ['start', 'end', 'gaps in days']

        gaps_fn = '{}_gaps.csv'.format(name)

        gaps.to_csv(gaps_fn, float_format='%0.3f', **csv_args(unix))

        print('The gaps table is saved {}.'.format(output_filename(gaps_fn)))

    # __________________________________________________________________________________________________________________
    if statistics:
        if tags is None:
            tags = data_validation(series)
        if availability is None:
            availability = data_availability(tags)

        availability_cut = 0.9

        stats = create_statistics(series, availability, availability_cut=availability_cut)

        with open('{}_stats.txt'.format(name), 'w+') as f:
            rain_fmt = '{:0.0f} mm'
            date_fmt = '{:%Y}'
            avail_fmt = '{:0.0%}'

            f.write(
                'The annual totals of the data series serve as the data basis.\n'
                'The following statistics were analyzed:\n'
                'Only years with a availability of {avail} will be evaluated.\n'
                '\n'
                'The maximum is {rain} and was in the year {date} (with {avail} Data available).\n'
                'The minimum is {rain} and was in the year {date} (with {avail} Data available).\n'
                'The mean is {rain} (with {avail} Data available in average).'
                ''.format(rain=rain_fmt, date=date_fmt, avail=avail_fmt).format(availability_cut,
                                                                                *stats['maximum'],
                                                                                *stats['minimum'],
                                                                                *stats['mean'])
            )
            print('The statistics are saved {}.'.format(output_filename(f.name)))

    # __________________________________________________________________________________________________________________
    save_formats = list()
    if to_csv:
        save_formats.append('csv')
    if to_parquet:
        save_formats.append('parquet')

    for save_as in save_formats:
        out_fn = export_series(series, filename=name, save_as=save_as, unix=unix)
        print('The time-series is saved {}.'.format(output_filename(out_fn)))

    # __________________________________________________________________________________________________________________
    if plot:
        if tags is None:
            tags = data_validation(series)
        if availability is None:
            availability = data_availability(tags)

        plot_fn = '{}_plot.png'.format(name)
        rain_plot(series, availability, plot_fn)
        print('The plot is saved {}.'.format(output_filename(plot_fn)))
        show = True
        if show:
            show_file(plot_fn)



# coding=UTF-8


def ehyd_reader(filename, output_type, write_csv='False',
                input_type='csv', interpolate='False'):
    """
    Reads in a CSV file containing a hydrologic time series (filename).

    Meant to be used with files obtained from ehyd.gv.at, taking into
    account the intricacies of this data (ISO 8859-15 file, use of
    german Umlauts, using the term "Lücke" for NaN, decimal comma).

    Can also read in .dat files, which get used internally in the
    governement agencies. Those follow the same format as the csv files
    but use a decimal dot and variable spaces as delimiter. Note that
    these files can have very irregular measurements, so handling these
    is still experimental!

    Use 'dict' or 'df' to specify wether you want a dict or dataframe as
    output.

    If you want a "proper" CSV file in your working directory, use
    "write_csv=True".

    Set interpolate to 'True' if you want to interpolate over missing
    data. For small gaps, this should be OK, but if 'data_error'
    indicates a lot of missing data, this is probably going to cause
    issues when further handling the data.

    Parameters
    ----------
    filename : str
        path to and name of the file to be read
    output_type : str
        can be 'df' for a dataframe as output or 'dict" for a dictionary
    write_csv : bool
    input_type : str
        can be 'csv' for a CSV file or 'dat' for a DAT file
    interpolate : bool

    Examples
    --------
    # Read in a csv file and output a dataframe:
    ehyd_df = ehyd_reader('filename.csv', output_type='df')
    # Read in a dat file and output a dictionary and a csv file:
    ehyd_dict = ehyd_reader('filename.dat', output_type='dict',
                            input_type='dat', write_csv='True')

    """

    # Import the needed modules.
    import pandas as pd
    import numpy as np
    import re
    from sklearn.metrics.pairwise import haversine_distances
    from math import radians
    # The last two imports are needed to figure out how much the station moved
    # over its lifetime. If that's not needed, the whole stuff around the
    # variable "station_move" can be removed.

    # A lot of metadata start with ';' and end with a newline.
    # Lets define a little function for that.
    def csv_splitter(tester):
        start_str = tester.find(';') + 1
        end_str = tester.find('\n')
        t_str = tester[start_str:end_str]
        return t_str

    def dat_splitter(tester):
        start_str = tester.find(':') + 1
        end_str = tester.find('\n')
        t_str = tester[start_str:end_str].lstrip()
        return t_str

    if input_type == 'dat':
        splitter = dat_splitter
    else:
        splitter = csv_splitter

    # define a long-ish function to read the coordinates,
    # needed twice, later
    def coord_calculator(raw_coords):
        nonlocal meta_error
        coord_string = str(raw_coords)
        if input_type == 'dat':
            coord_list = re.findall(r"\d{2}\s\d{2}\s\d{2}", coord_string)
            raw_longitude = coord_list[0]
            raw_latitude = coord_list[1]
        else:
            longitude_start = coord_string.find(';') + 1
            longitude_end = coord_string.rfind(';') - 1
            latitude_start = coord_string.rfind(';') + 1
            latitude_end = coord_string.find('\n')
            raw_longitude = coord_string[longitude_start:longitude_end]
            raw_latitude = coord_string[latitude_start:latitude_end]
        # Remove trailing whitespace.
        stripped_lon = raw_longitude.rstrip()
        stripped_lat = raw_latitude.rstrip()
        # There can be cases where lat and lon are missing i.e. meaning they're
        # set to 0. If that's the case, all the stuff below fails. so we gotta
        # test for that.
        if stripped_lat == '0':
            lat = np.nan
            lon = np.nan
            print('Station', HZB, 'is missing coordinates!')
            meta_error = 'missing_coords'
        else:
            # Grab the parts and turn them into floats
            s_deg_lon = stripped_lon[0:2]
            deg_lon = float(s_deg_lon)
            s_min_lon = stripped_lon[3:5]
            min_lon = float(s_min_lon)
            s_sec_lon = stripped_lon[6:8]
            sec_lon = float(s_sec_lon)
            s_deg_lat = stripped_lat[0:2]
            deg_lat = float(s_deg_lat)
            s_min_lat = stripped_lat[3:5]
            min_lat = float(s_min_lat)
            s_sec_lat = stripped_lat[6:8]
            sec_lat = float(s_sec_lat)
            # Put it all together
            lon = deg_lon + (min_lon / 60) + (sec_lon / 3600)
            lat = deg_lat + (min_lat / 60) + (sec_lat / 3600)
            lon = round(lon, 8)
            lat = round(lat, 8)
            # Should only result in 6 decimals,
            # but the division could result in a repeating decimal,
            # so let's keep that at a sane size.
        return lat, lon

    # since filename is kinda generic, change it to current_csv
    current_csv = filename

    f1 = open(current_csv, 'r', encoding='cp1252')
    table = f1.readlines(3000)
    # Reads in only 3000 characters/bytes. Most ehyd files appear to contain
    # around 1500 characters in the metadata, so this has a good margin of
    # error. Since readlines is quite fast, this limitation can also be
    # discarded without much of a penalty( ~100µs vs. 9ms for a 1MB file).
    f1.close()

    # Extract the name of the station:
    station_name_line = table[0]
    station_name = splitter(station_name_line)

    # Set metadata that might not be in every file to the 'NaN' string, to show
    # that they're not available.
    province = 'NaN'
    catchment_size = 'NaN'
    catchment_symbol = 'NaN'
    station_move = 'NaN'
    subcatchment = 'NaN'
    region_name = 'NaN'
    # set depth to not available for things that do not have it
    depth = 'NaN'
    elev = 'NaN'

    # Grab various important information from the header and find the line in
    # which the data starts.
    # Will fail when a file is smaller than 40 lines, but such short time
    # series do not make sense to work with.
    # Will also fail if the headers of the csvs get bigger than 40 lines.
    # As of now, they're around 34, so 40 leaves a bit of a margin.
    # If it fails, increase header_counter as you see fit.
    # AS of now, this can only handle groundwater levels, river levels,
    # precipitation and spring flows. If it shall be extended to other types,
    # some logic is needed to distinguish between the different possibilities
    # that one of the current 'datatype' can contain. E.g. river temperature
    # not only needs to search for 'Gew\xe4sser' as of now, but also for
    # 'WTemperatur'.
    # Thus, we set the datatype to NA at first, and change it only for files
    # that fit the rest of the Analysis:
    datatype = 'NA'
    meta_error = 'no_error'
    data_error = 'no_error'
    header_counter = 0
    while header_counter < 40:
        tester_list = table[header_counter]
        tester = str(tester_list)  # must be turned into a string,
        # otherwise we cant find strings in it.
        if 'Gew\xe4sser' in tester:
            datatype = 'Riverwater'
            catchment_name = splitter(tester)
        if 'Niederschlag' in tester:
            datatype = 'Precipitation'
            catchment_name = station_name
            # Catchment name is set to the name of the station.
        if 'Hauptquelle' in tester:
            datatype = 'Spring'
        if 'Grundwasser' in tester:
            datatype = 'Groundwater'
            catchment_name = splitter(tester)
            # Gets the written name and short symbol for the catchment, e.g.
            # 'Grazer Feld (Graz/Andritz - Wildon) [MUR]'
            # the Murdurchbruchstal is missing the [MUR] marker.
            # So we have to deal with that too.
            if '[' in catchment_name:
                name_end = catchment_name.find(']')
                name_start = name_end - 3
                catchment_symbol = catchment_name[name_start:name_end]
            else:
                catchment_symbol = 'NA'
            # Gets the short symbol for the catchment, such as
            # MUR, DRA (=Drau), DUJ (=Danube below Jochenstein) etc
            groundwaterbody_name = table[header_counter - 1]
            # Normally there are two names for the groundwater body, the larger
            # 'Grundwasserkörper' and the subregion 'PorenGW-Gebiet' which can
            # be very local (few square km).
            subcatchment = splitter(groundwaterbody_name)
        if 'Dienststelle' in tester:
            region_name = splitter(tester)
            # Who is responsible for the station. Mostly indicates which state.
            # E.g. 'HD-Steiermark'. Can overlap with 'Messstellenbetreiber' and
            # 'Bundesland', the latter being rare, however.
            # For Vienna and Lower Austria, it can also contain
            # 'HD-Wien (MA 45)' or 'Magistratsabteilung 31' with added clutter
            # or no indication of the state.
            # In this case there can also be a 'Bundesland' row:
        if 'Bundesland' in tester:
            province = splitter(tester)
            # And to complicate matters further, there can also be a
            # 'Messstellenbetreiber', i.e. who operates (or owns?) the station:
        if 'Messstellenbetreiber' in tester:
            operator = splitter(tester)
        if 'HZB-Nummer' in tester:
            # Keep the unique identifier used in ehyd
            HZB = int(splitter(tester))
            print('HZB:', HZB)
        if 'HD-Nummer' in tester:
            # Besides the HZB number, there's often also the HD number, used by
            # the state authorities. Often, this number is also included in the
            # name of the station. Can also contain letters and spaces, so
            # we'll leave that as a string.
            HD_num = splitter(tester)
        if 'DBMS-Nummer' in tester:
            # Some stations also have a DBMS number besides or instead of the
            # HD number. Seems like it's only numbers, but nothing known about
            # it, so lets keep it string.
            DBMS_num = splitter(tester)
        if 'Einzugsgebiet' in tester:
            # find the size of the catchment
            catchment_size = splitter(tester)
        if 'Koordinaten' in tester:
            # Sometimes a station gets moved and there will be all the
            # coordinates it ever had listed in the file. The lines below try
            # to find the last (=most current) coordinates by searching for the
            # term 'Exportzeitreihe' which comes after the last location.
            coord_tester = table[header_counter + 3]
            if 'Exportzeitreihe' in coord_tester:
                # This means there is only one row of Coordinates, so the
                # station didn't move.
                coord_values = header_counter + 2
                coords_changed = 'False'
            else:
                # There's more than one row of Coordinates. We need to set the
                # first one as old_coords to calculate the distance.
                print('Station got moved during its livetime.')
                coords_changed = 'True'
                meta_error = 'Station_moved'
                old_coords = table[header_counter + 2]
                old_coords = coord_calculator(old_coords)
                coord_tester = table[header_counter + 4]
                if 'Exportzeitreihe' in coord_tester:
                    coord_values = header_counter + 3
                else:
                    coord_tester = table[header_counter + 5]
                    if 'Exportzeitreihe' in coord_tester:
                        coord_values = header_counter + 4
                    else:
                        coord_values = header_counter + 5
                        print('Station', HZB, 'has been moved at least 4 times! \n',
                              'Coordinates might be of an older location, please check this station!')

            coords = coord_calculator(table[coord_values])
            lat = coords[0]
            lon = coords[1]
            if coords_changed == 'True':
                # Print info about how far the station got moved. Sometimes,
                # coordinates are missing, so we need some routine to deal with
                # missing coordinates
                if meta_error == 'missing_coords':
                    print('Station', HZB, 'was moved by an unknown distance! \n',
                          'Using the coordinates of the older location!')
                    coords = old_coords
                    lat = coords[0]
                    lon = coords[1]
                else:
                    # Uses the haversine distance from scikit learn. see
                    # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.haversine_distances.html
                    # or
                    # https://en.wikipedia.org/wiki/Haversine_formula
                    # could probably also be implemented without scikit-learn.
                    # Needs to be turned into radians to work with.
                    coords_rad = [radians(i) for i in coords]
                    old_coords_rad = [radians(i) for i in old_coords]
                    station_move = haversine_distances(
                        [coords_rad, old_coords_rad]) * 6371 * 1000
                    # Multiply with earth radius in meters to get the distance.
                    station_move = round(station_move[0, 1], 2)
                    # output is an array, this should get the distance without
                    # the rest, rounded to two decimals (=centimeters).
                    print('Sation', HZB, 'was moved by',
                          int(station_move), 'meters.')

        if 'Messpunkthöhe:' in tester:
            elev = tester.replace(",", ".")
            elev = splitter(elev)
            elev = float(elev)
        if 'Sohllage' in tester:
            depth = tester.replace(",", ".")
            depth = splitter(depth)
            depth = float(depth)
        if 'Geländehöhe-Hauptquelle' in tester:
            # can be an elevation of a spring.
            # but is this really needed? What about Pegelnullpunk?
            elev = tester.replace(",", ".")
            elev = splitter(elev)
            if elev == '':
                elev = 'NaN'
            else:
                elev = float(elev)
        if 'Pegelnullpunkt' in tester and datatype == 'Spring' and elev == 'NaN':
            elev = tester.replace(",", ".")
            elev = splitter(elev)
            elev = float(elev)
        if 'Pegelnullpunkt' in tester and datatype == 'Riverwater':
            elev = str(table[header_counter + 2])
            elev = elev.replace(",", ".")
            elev = splitter(elev)
            if elev == '':
                elev = 'NaN'
            else:
                elev = float(elev)
        if 'Höhe:' in tester:  # and datatype == 'Precipitation':
            # Should also test for datatype, but this fails.
            # Seems like 'Höhe:' only gets used in precip, but a test would
            # still be nice.
            elev = str(table[header_counter + 2])
            elev = elev.replace(",", ".")
            elev = splitter(elev)
            if elev == '':
                elev = 'NaN'
            else:
                elev = float(elev)
        # Can also have a lot of different values if the station moved.
        # Still missing something to detect that and to give info about changes
        # station location. Uses the first value it finds. Will probably not
        # be totally wrong, but isn't the most recent value.
        if 'Werte:' in tester:
            werte_start = header_counter
            # Finds the line where the data starts
        header_counter += 1

    if depth != 'NaN':
        teufe = round((elev - depth), 2)
        # Due to how floating point numbers work, we sometimes run into values
        # like 11.799999999999983 where it should be 11.8 so lets round it to
        # take care of that.
    else:
        teufe = 'NaN'

    if input_type == 'dat':
        seperator = '\s{2,10}'
        deci = '.'
    else:
        seperator = '\s*\;\s*'
        deci = ','

    hydro_tS = pd.read_csv(current_csv,
                           skiprows=werte_start + 1, decimal=deci,
                           sep=seperator, usecols=[0, 1], index_col=0,
                           dayfirst=True, parse_dates=True, engine='python',
                           encoding='cp1252', names=['date', 'level'],
                           na_values='Lücke')
    # Due to the weird format of the csv files, they appear to be enconded as
    # cp1252. If ehyd gets updated at some point in time or this is adapeted
    # to other files, 'encoding='cp1252'' might have to be changed to some more
    # common format.

    # The data sometimes gets read as object, but we need float.
    hydro_tS.level = hydro_tS.level.astype(float)

    # Normalize the whole thing to get rid of hours and minutes.
    # If those are needed, the following line can be removed,
    # but it'll wreak havok with irregular times.
    hydro_tS.index = hydro_tS.index.normalize()

    # get the lowest and the largest date
    min_date = hydro_tS.index.min()
    max_date = hydro_tS.index.max()
    mini_date_1 = str(min_date)
    maxi_date_1 = str(max_date)
    mini_date = mini_date_1[0:10]
    maxi_date = maxi_date_1[0:10]
    # hardcoded to get rid of the hours and seconds.
    # Might fail if anything in the date format changes.

    min_year = min_date.year
    max_year = max_date.year
    data_years = max_year - min_year

    # Check if there is a gap in the index, which can cause issues with some
    # further procsessing. First we need to find the frequency of the data.
    # We also need to get rid of the hours of the day, else days with multiple
    # measurements will wreak havoc.
    firstdate = hydro_tS.index[0]
    seconddate = hydro_tS.index[1]
    datedelta = seconddate - firstdate
    # Pandas timedelta. Seems like it defaults to days, but just to make sure:
    dayinterval_1 = datedelta.days
    # Results in an int with the number of days between two days.
    # 1=daily timeseries; 28-31=monthly
    # However, some csv files can havechanging intervals. Let's do a few more
    # tests to find out, by probing the interval at the end and in the middle.
    lastdate = hydro_tS.index[-1]
    second_lastdate = hydro_tS.index[-2]
    datedelta = lastdate - second_lastdate
    dayinterval_2 = datedelta.days
    middate = hydro_tS.index[len(hydro_tS.index) // 2]
    smiddate = hydro_tS.index[len(hydro_tS.index) // 2 - 1]
    datedelta = middate - smiddate
    dayinterval_3 = datedelta.days

    if dayinterval_1 == dayinterval_2 == dayinterval_3 == 1:
        TS_freq = 'D'
        # hydro_tS = hydro_tS.asfreq(TS_freq)

    elif min(dayinterval_1, dayinterval_2, dayinterval_3) > 1:
        if min(dayinterval_1, dayinterval_2, dayinterval_3) > 27:
            TS_freq = 'MS'
            # hydro_tS = hydro_tS.asfreq(TS_freq)
        else:
            TS_freq = 'MS'
            # hydro_tS = hydro_tS.asfreq(TS_freq)
            print('Station', HZB, 'has irregular monthly times! \n',
                  'Check the index before doing calculations on it.')
            data_error = 'Irregular_measure_times_M'

    else:
        TS_freq = 'D'
        if input_type == 'dat':
            hydro_tS = hydro_tS.resample(TS_freq).mean()
            print('Station', HZB, 'has multiples measurements per day. \n',
                  'Resampled to daily.')
        else:
            print('Station', HZB, 'has irregular measurement times! \n',
                  'Check the index before doing calculations on it.')
        data_error = 'Irregular_measure_times_D'
    # The commented out
    # hydro_tS = hydro_tS.asfreq(TS_freq)
    # lines can resample it to the frequency it apparently has.
    # Will pad missing dates and insert NaN as a value,
    # which will then be taken care of by the stuff below.

    # Built a empty date range with the same start, end and frequency as the
    # orignal data and get its length:
    test_dates = pd.date_range(mini_date, maxi_date, freq=TS_freq)
    test_length = len(test_dates)

    # Get the length of the real data and compare it. If it differs, there must
    # be a gap in the data.
    real_length = len(hydro_tS)
    # Counts everything, including NaN.
    lengthdifference = test_length - real_length

    if lengthdifference != 0:
        if input_type == 'csv':
            hydro_tS = hydro_tS[~hydro_tS.index.duplicated()]
            # Get rid of duplicate entries, keeping the first one. For data that
            # has many measurements per day, this is not a good idea. ehyd data
            # should have only one per day, so this deals with some rare errors,
            # but for the raw data, this can be problematic.
            # the ~ is apparently a reversal of the booleans in that duplicated.
            # see
            # https://stackoverflow.com/questions/13035764/remove-rows-with-duplicate-indices-pandas-dataframe-and-timeseries
            print('Two measurements at one day. first measurement deleted')
        data_error = 'double_measurement'
    if lengthdifference > 4:
        # Apparently, there is some issues around 1945 (war?) where total
        # months are missing, including dates.
        data_error = 'gap_data'
        print('there is a gap in the data')
        print('lengthdifference')
        print(lengthdifference)
    if data_years < 5:
        # Data that is too short can't reliably be used for many statistics.
        data_error = 'short_data'
        print('time series very short')
    if min_year > 2010:
        data_error = 'short_data'
        print('time series very short')
    else:
        # count the NaNs:
        real_count = hydro_tS.level.count()  # counts only valid numbers
        NaN_number = real_length - real_count
        NaN_number = float(NaN_number)
        percent_data = real_length / 100
        ten_percent = percent_data * 10
        if NaN_number > ten_percent:
            data_error = '10+percent_gap'
            print('10 percent of data missing')

        # Count how many NaNs there are following each other
        consecutive_NaNs = hydro_tS.level.isna().astype(int).groupby(
            hydro_tS.level.notna().astype(int).cumsum()).sum().max()

        if TS_freq == 'D' and consecutive_NaNs > 14:
            data_error = '14+dayGap'
            print('more than 14 consecutive NaN in daily data')
        if TS_freq == 'MS' and consecutive_NaNs > 3:
            data_error = '3+monthsGap'
            print('more than 3 consecutive NaN in monthly data')

        if interpolate == 'True':
            # Get rid of the NaN in the dataset.
            # This is just a crude interpolation.
            # For data with just a few, this shouldn't be an issue, but if
            # data_error = '10+percent_gap', '3+monthsGap' or '14+dayGap'
            # this is probably not a valid approach.
            hydro_tS.level = hydro_tS.level.interpolate()

    if output_type == 'df' or write_csv == 'True':
        output_header = pd.MultiIndex.from_product(
            [[datatype], [station_name], [HZB], [HD_num],
             [DBMS_num], [catchment_name], [catchment_symbol],
             [subcatchment], [region_name], [province], [operator],
             [catchment_size], [lat], [lon], [station_move],
             [elev], [depth], [teufe], [data_error], [meta_error]],
            names=['datatype', 'station_name', 'HZB', 'HD_num',
                   'DBMS_num', 'catchment_name', 'catchment_symbol',
                   'subcatchment', 'region_name', 'province', 'operator',
                   'catchment_size', 'lat', 'lon', 'station_move',
                   'elev', 'depth', 'teufe', 'data_error', 'meta_error'])
        hydro_tS.columns = output_header
        output = hydro_tS
    if write_csv == 'True':
        filename_out = str('%s.csv' % (HZB))
        hydro_tS.to_csv(filename_out, sep=';')

    if output_type == 'dict':
        output = {'data_error': data_error, 'meta_error': meta_error,
                  'station_name': station_name, 'catchment_name': catchment_name,
                  'catchment_symbol': catchment_symbol,
                  'subcatchment': subcatchment, 'region_name': region_name, 'HZB': HZB,
                  'HD_num': HD_num, 'DBMS_num': DBMS_num, 'province': province,
                  'operator': operator, 'catchment_size': catchment_size,
                  'lat': lat, 'lon': lon, 'elev': elev, 'depth': depth,
                  'teufe': teufe, 'station_move': station_move,
                  'datatype': datatype, 'timeseries': hydro_tS}

    return output



import numpy as np
import pandas as pd
from math import floor
from abc import ABC, abstractmethod

from scipy.interpolate import interp2d


def block_rain(idf_table, return_period, duration, interval=5):
    if duration not in idf_table.index:
        idf_table = idf_table.append(pd.Series(index=idf_table.columns, name=duration)).sort_index()

    if isinstance(return_period, float):
        idf_table.columns = idf_table.columns.astype(float)
    if return_period not in idf_table.columns:
        idf_table[return_period] = np.NaN
        idf_table = idf_table.sort_index(axis=1)

    if any(idf_table.isna()):
        idf_table = idf_table.interpolate().T.interpolate().T

    height = idf_table.loc[duration, return_period]
    index = range(interval, duration + interval, interval)
    intensity = height / len(index)
    r = pd.Series(index=index, data=intensity)

    r = r.append(pd.Series({0: 0})).sort_index()
    return r


def euler_model_rain(idf_table, return_period, duration, interval=5, occurrence_highest_intensity=1 / 3):
    return_period_series = idf_table[return_period]

    index = range(interval, duration + interval, interval)
    filtered_series = pd.Series(data=np.interp(index, return_period_series.index, return_period_series),
                                index=index)

    d = filtered_series.diff()
    d.iloc[0] = filtered_series.iloc[0]

    # max_index = floor((occurrence_highest_intensity * duration) / interval) * interval
    max_index = floor(round((occurrence_highest_intensity * duration) / interval, 3)) * interval

    # sort differences and reset index
    r = d.sort_values(ascending=False)
    r.index = sorted(r.index.values)

    # reverse first <occurrence_highest_intensity> values
    r.loc[:max_index] = r.loc[max_index::-1].values

    # add Zero value at first posiotion (for SWMM ?)
    r = r.append(pd.Series({0: 0})).sort_index()

    # ---------------
    if 0:
        _res = pd.DataFrame(index=filtered_series.index)
        _res['basis'] = return_period_series
        _res['gefüllt'] = filtered_series
        _res['diff'] = d
        _res['spitze'] = False
        _res.loc[max_index, 'spitze'] = True
        _res['fertig'] = r
    return r


def euler_model_rain_type2(idf_table, return_period, duration, interval=5):
    return euler_model_rain(idf_table, return_period, duration, interval=interval, occurrence_highest_intensity=1 / 3)


def euler_model_rain_type1(idf_table, return_period, duration, interval=5):
    return euler_model_rain(idf_table, return_period, duration, interval=interval, occurrence_highest_intensity=0)


def rain_series(start_time, kind, idf_table, return_period, duration, interval=5):
    rain = kind(idf_table, return_period=return_period, duration=duration, interval=interval)
    rain.index = start_time + pd.to_timedelta(rain.index, unit='m')
    return rain


if __name__ == '__main__':
    from ehyd_tools.design_rainfall import get_ehyd_design_rainfall_file, read_ehyd_design_rainfall, get_max_calculation_method

    grid_point = 5214
    idf_table = read_ehyd_design_rainfall(get_ehyd_design_rainfall_file(grid_point_number=grid_point))
    idf_table = get_max_calculation_method(idf_table)
    print(block_rain(idf_table, return_period=2, duration=60, interval=5))
    print(euler_model_rain_type2(idf_table, return_period=2, duration=60, interval=5))
    print(euler_model_rain_type2(idf_table, return_period=3, duration=60, interval=5))



__author__ = "David Camhy, Markus Pichler"
__copyright__ = "Copyright 2017, University of Technology Graz"
__credits__ = ["David Camhy", "Markus Pichler"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "David Camhy, Markus Pichler"

from pandas import Timedelta, DataFrame, isna, Series, DateOffset
from numpy import NaN
from warnings import warn
from matplotlib.pyplot import subplots


def start_end_date(series):
    return series.index[[0, -1]].tolist()


def year_delta(years):
    """
    return a timedelta object for a given number of years

    Args:
        years (float): number of years

    Returns:
        pandas.Timedelta: time period
    """
    return Timedelta(days=365.25) * years


def data_validation(series):
    """
    add gaps sum & nan sum to the time series

    Args:
        series (pandas.Series): time-series

    Returns:
        pandas.DataFrame: tags with columns 'nans', 'gaps', ...
    """
    ts = series.copy()

    first_index = ts.index[0].replace(day=1, month=1, hour=0, minute=0)
    if first_index not in series.index:
        ts = Series(index=[first_index]).append(ts)

    last_index = ts.index[-1].replace(day=31, month=12, hour=23, minute=59)
    if last_index not in ts.index:
        ts = ts.append(Series(index=[last_index]))

    if ts.index.has_duplicates:  # very slow an large data sets
        ts = ts[~ts.index.duplicated()].copy()

    if not ts.index.is_monotonic_increasing:
        raise UserWarning('Series has not monotonic increasing of the timestamps.')
        ts = ts.sort_index()

    tags = DataFrame(index=ts.index)
    tags['nans'] = isna(ts).astype(int)
    tags = tags.reindex(tags.asfreq('T').index)
    tags['gaps'] = isna(ts.fillna(0).reindex(tags.index)).astype(int)
    return tags


def data_availability(tags):
    """
    get availability based on the validation tags

    Args:
        tags (pandas.DataFrame): errors tagged as true (see function data_validation)

    Returns:
        pandas.Series: availability
    """
    return ~tags.any(axis=1)


def max_10a(availability):
    """
    get the minimal gaps and minimal nan
    and take the latest possible time span

    Args:
        availability (pandas.Series[bool]): series with available data tagged as True

    Returns:
        tuple[pandas.Timestamp, pandas.Timestamp]: start & end time of most available time span
    """

    window = year_delta(years=10)
    avail_sum = availability.rolling(window).sum()

    end = avail_sum[::-1].idxmax()
    start = end - DateOffset(years=10)
    return start, end


def check_period(series):
    """
    check if the period of the series is at least 10 years long

    Warnings:
        UserWarning - if the series is shorter

    Args:
        series (pandas.Series): time-series
    """
    if not is_longer(series, years=10):
        print('WARNING: The Series is only {:.1f} < 10 years long!'.format(
            (series.index[-1] - series.index[0]) / Timedelta(days=365)))


def is_longer(series, years):
    """
    check if the period of the series is at least x years long

    Args:
        series (pandas.Series): time-series
        years (int): number of years to compare with

    Returns:
        bool: whether the series is longer
    """
    return series.index[0] + DateOffset(years=years) <= series.index[-1]
    # return (series.index[-1] - series.index[0]) > year_delta(years=years)


def rain_figure(series, availability, freq=None, add_mean_line=False):
    """
    creates a monthly sum bar plot

    Args:
        series (pandas.Series): time-series
        availability (pandas.Series): availability
        fn (str): path + filename of the resulting plot
    """

    freq_long = {
        'Y': 'Jahr',
        'M': 'Monat'
    }

    if freq is None:
        if is_longer(series, years=15):
            freq = 'Y'
            # base_delta = year_delta(1)
        else:
            freq = 'M'
            # base_delta = year_delta(1) / 12

    sums = series.resample(freq).sum()

    if freq == 'Y':
        index = sums.index.year.values
    else:
        # monthly
        index = sums.index.year.values + sums.index.month.values / 12

    # index = series.index.to_series().resample(freq).apply(lambda s: s.min() + abs(s.min() - s.max()) / 2).dt.date.values
    freq_avail = availability.resample(freq)
    avail = (freq_avail.sum() / freq_avail.count())[sums.index]

    # dummy = sums.rename('dummy').copy()
    # dummy.loc[:] = 0

    height_ratios = [1, 10]

    fig, (ax1, ax) = subplots(2, gridspec_kw=dict(height_ratios=height_ratios), sharex=True)

    # ax = dummy.plot(ax=ax, lw=0)
    fig.set_size_inches(w=29.7 / 2.54, h=21. / 2.54)

    ax1.set_ylim(0, 100)
    ax1.set_ylabel('% Verfügbar')
    ax1.set_yticks(range(0, 110, 10), minor=True)
    ax1.grid(axis='y', which='minor', color='lightgrey', linestyle=':', linewidth=0.5)  # , zorder=-10000000)
    # ax1 = dummy.plot(ax=ax1, lw=0)
    # ax1.scatter(x=index, y=avail.values * 100, color='grey', clip_on=False, marker='_')
    ax1.bar(x=index, height=avail.values * 100, color='grey')
    ax1.set_axisbelow(True)

    if add_mean_line:
        mean = sums[avail > 0.5].mean()
        ax.text(ax.get_xlim()[0] + 1, mean + 15, f'Mittelwert: {mean:0.0f} (mm/{freq_long[freq]})',
                bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 2, 'linewidth': '0'})
        ax.axhline(mean, ls='--', color='darkgray', linewidth=0.7)

    ax.bar(x=index, height=sums.values, color='k')
    ax.set_ylabel('Niederschlag (mm/{})'.format(freq_long[freq]))
    ax.set_xlabel('Zeit')
    # ax.set_xlim(left=ax.get_xlim()[0] - 0.5)
    # ax.set_xlim(right=ax.get_xlim()[1] + 0.5)
    fig.tight_layout()
    fig.subplots_adjust(hspace=0)
    return fig, ax


def rain_plot(series, availability, fn):
    """
    creates a monthly sum bar plot

    Args:
        series (pandas.Series): time-series
        availability (pandas.Series): availability
        fn (str): path + filename of the resulting plot
    """
    fig, ax = rain_figure(series, availability)
    # fig.show()
    fig.savefig(fn)


def create_statistics(series, availability, availability_cut=0.2):
    """
    creates basic (maximum, minimum and mean) yearly statistics of the time-series

    Args:
        series (pandas.Series): time-series
        availability (pandas.Series): availability
        availability_cut (float): minimal availability for creating a yearly statistic

    Returns:
        dict: of the yearly statistics
    """
    sums = series.resample('Y').sum()
    avail_sum = availability.resample('Y').sum()

    base_freq = series.index.freq
    yearly_index = avail_sum.index
    delta_per_year = yearly_index - (yearly_index - DateOffset(years=1))
    avail_full = delta_per_year / base_freq
    avail = avail_sum / avail_full  # type: Series
    if (avail < availability_cut).all():
        print('ATTENTION: only very small data availability! The statistic may be not very meaningful.')
        if (avail < 0.1).all():
            return dict()
        sums[avail < 0.1] = NaN
    else:
        sums[avail < availability_cut] = NaN

    stats = dict()
    stats['maximum'] = (sums.max(), sums.idxmax(), avail.loc[sums.idxmax()])
    stats['minimum'] = (sums.min(), sums.idxmin(), avail.loc[sums.idxmin()])
    stats['mean'] = (sums.mean(), avail.mean())
    return stats



import numpy as np
import pandas as pd
from math import floor
from abc import ABC, abstractmethod

from scipy.interpolate import interp2d

from ehyd_tools.design_rainfall import (get_ehyd_design_rainfall_file, read_ehyd_design_rainfall,
                                        get_max_calculation_method, get_calculation_method, )


class _AbstractModelRain(ABC):
    def __init__(self, idf_table=None):
        self.idf_table = idf_table

    def _get_index(self, duration, interval=5):
        return range(interval, duration + interval, interval)

    def _get_idf_value(self, duration, return_period):
        f = interp2d(x=self.idf_table.columns.values, y=self.idf_table.index.values, z=self.idf_table.values,
                     kind='linear')
        return float(f(return_period, duration)[0])

    @abstractmethod
    def _get_series(self, return_period, duration, interval=5, **kwargs):
        pass

    def get_time_series(self, return_period, duration, interval=5, start_time=None, **kwargs):
        rain = self._get_series(return_period, duration, interval, **kwargs)
        if start_time is not None:
            if isinstance(start_time, str):
                start_time = pd.to_datetime(start_time)
            rain.index = start_time + pd.to_timedelta(rain.index, unit='m')
        return rain


class _BlockRain(_AbstractModelRain):
    def __init__(self, idf_table=None):
        _AbstractModelRain.__init__(self, idf_table)

    def _get_series(self, return_period, duration, interval=5, **kwargs):
        height = self._get_idf_value(duration, return_period)
        index = self._get_index(duration, interval)
        intensity = height / len(index)
        r = pd.Series(index=index, data=intensity)
        r = r.append(pd.Series({0: 0})).sort_index()
        return r


class _EulerRain(_AbstractModelRain):
    def __init__(self, idf_table=None):
        _AbstractModelRain.__init__(self, idf_table)

    @staticmethod
    def _get_occurrence_highest_intensity(kind=2):
        if kind == 1:
            return 0
        elif kind == 2:
            return 1 / 3

    def _get_series(self, return_period, duration, interval=5, kind=2):
        return_period_series = self.idf_table[return_period]

        index = self._get_index(duration, interval)
        filtered_series = pd.Series(data=np.interp(index, return_period_series.index, return_period_series),
                                    index=index)

        d = filtered_series.diff()
        d.iloc[0] = filtered_series.iloc[0]

        # max_index = floor((occurrence_highest_intensity * duration) / interval) * interval
        max_index = floor(round((self._get_occurrence_highest_intensity(kind) * duration) / interval, 3)) * interval

        # sort differences and reset index
        r = d.sort_values(ascending=False)
        r.index = sorted(r.index.values)

        # reverse first <occurrence_highest_intensity> values
        r.loc[:max_index] = r.loc[max_index::-1].values

        # add Zero value at first posiotion (for SWMM ?)
        r = r.append(pd.Series({0: 0})).sort_index()

        # ---------------
        if 0:
            _res = pd.DataFrame(index=filtered_series.index)
            _res['basis'] = return_period_series
            _res['gefüllt'] = filtered_series
            _res['diff'] = d
            _res['spitze'] = False
            _res.loc[max_index, 'spitze'] = True
            _res['fertig'] = r
        return r


class RainModeller:
    def __init__(self, idf_table=None):
        """
        object to create block and euler model rain series

        Args:
            grid_point (int): 4 digit id number (Gitterpunktnummer)
            return_period (int): in years
            duration (int): in minutes
            interval (int): in minutes
            kind (str): which row to use of the idf table. one of ['ÖKOSTRA', 'Bemessung', 'maxBemessung']
        """
        self.idf_table = idf_table

    def set_idf_table_okostra(self, grid_point, kind='maxBemessung'):
        """
        set idf table of ehyd database

        Args:
            grid_point (int): 4 digit id number (Gitterpunktnummer)
            kind (str): which row to use. one of ['ÖKOSTRA', 'Bemessung', 'maxBemessung']
        """
        self.idf_table = read_ehyd_design_rainfall(get_ehyd_design_rainfall_file(grid_point_number=grid_point))
        if kind == 'maxBemessung':
            self.idf_table = get_max_calculation_method(self.idf_table)
        else:
            self.idf_table = get_calculation_method(self.idf_table, method=kind)

    @property
    def block(self) -> _BlockRain:
        return _BlockRain(idf_table=self.idf_table)

    @property
    def euler(self, ) -> _EulerRain:
        return _EulerRain(idf_table=self.idf_table)


if __name__ == '__main__':
    model_rain = RainModeller()
    model_rain.idf_table = pd.read_csv('/home/markus/PycharmProjects/ehyd_tools/example/kostra_s24z70.csv', index_col=0)
    model_rain.idf_table.columns = model_rain.idf_table.columns.astype(int)
    print(model_rain.block.get_time_series(return_period=2, duration=60, interval=5))
    exit()
    model_rain = RainModeller()
    model_rain.set_idf_table_okostra(5214, kind='Bemessung')
    print(model_rain.block.get_time_series(return_period=2, duration=60, interval=5))

    model_rain2 = RainModeller()
    model_rain2.set_idf_table_okostra(5214, kind='maxBemessung')
    print(model_rain2.block.get_time_series(return_period=2, duration=60, interval=5))
    print(model_rain2.euler.get_time_series(return_period=2, duration=60, interval=5, kind=2))
    print(model_rain2.euler.get_time_series(return_period=2, duration=60, interval=5, kind=2, start_time='2021-05-27 16:00'))
